#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/iofunc.h>
#include <sys/dispatch.h>
#include <devctl.h>
#include <stdint.h>
#include <sys/mman.h>
#include <hw/inout.h>
#include <signal.h>
#include <sched.h>

#include "hitec.h"

#if 0
#define IRQ0		0
#define IRQ0_FREQ	100000
#define PWM_FREQ	50

#define PWM_CYCLE	(IRQ0_FREQ/PWM_FREQ)
#define PWM_MIN		60
#define PWM_MAX		260
#define PWM_MOD		(PWM_MAX-PWM_MIN)
#endif
#if 1
#define IRQ0		0
#define IRQ0_FREQ	50000
#define PWM_FREQ	50

#define PWM_CYCLE	(IRQ0_FREQ/PWM_FREQ)
#define PWM_MIN		30
#define PWM_MAX		120
#define PWM_MOD		(PWM_MAX-PWM_MIN)
#endif
#if 0
#define IRQ0		0
#define IRQ0_FREQ	10000
#define PWM_FREQ	50

#define PWM_CYCLE	(IRQ0_FREQ/PWM_FREQ)
#define PWM_MIN		6
#define PWM_MAX		24
#define PWM_MOD		(PWM_MAX-PWM_MIN)
#endif

unsigned char pwm_val[PWM_MOD];

#define LPT			0x378

static resmgr_connect_funcs_t connect_funcs;
static resmgr_io_funcs_t io_funcs;
static iofunc_attr_t attr;

const struct sigevent *
isr_handler (void *arg, int id);

int io_devctl (resmgr_context_t * ctp, io_devctl_t * msg, RESMGR_OCB_T * ocb);
data_t curr_pos;

struct _clockperiod oldcp, newcp;
int irq_id = -1;

void
sigint(int signum)
{
  if (irq_id != -1) {
  	if (InterruptDetach(irq_id) == -1) {
  		perror("InterruptDetach()");
  	} else {
#ifdef VARIANT_g  	
  		printf("ISR detached\n");
#endif  		
  	}
  }
  if (ClockPeriod (CLOCK_REALTIME, &oldcp, &newcp, 0) == -1) {
  	perror ("ClockPeriod()");
  } else {
#ifdef VARIANT_g  
		printf("ClockPeriod: new = %ldns, old = %ldns\n",
						oldcp.nsec, newcp.nsec);
#endif  	
  }
}

int
main (int argc, char **argv)
{
  /* declare variables we'll be using */
  resmgr_attr_t resmgr_attr;
  dispatch_t *dpp;
  dispatch_context_t *ctp;
  int id;
  unsigned char v;
  data_t initial_position;
  int i;
  pid_t pid;
  int policy;
  struct sched_param param;
  
  initial_position.tx_pos[0] = 42;
  initial_position.tx_pos[1] = 77;
  initial_position.tx_pos[2] = 55;  
  initial_position.tx_pos[3] = 29;
  initial_position.tx_pos[4] = 51;
  initial_position.tx_pos[5] = 37;
  initial_position.tx_pos[6] = 66;
  initial_position.tx_pos[7] = 81;
  
  pos2pwm(&initial_position);
  
  if (argc > 1) {
	  v = atoi(argv[1]);
  } else {
	  v = 0x00;
  }
  //printf("0x%02x\n", v);

  //memset(pwm_val, v, sizeof(pwm_val));

  /* initialize dispatch interface */
  if ((dpp = dispatch_create ()) == NULL)
    {
      fprintf (stderr, "%s: Unable to allocate dispatch handle.\n", argv[0]);
      return EXIT_FAILURE;
    }

  /* initialize resource manager attributes */
  memset (&resmgr_attr, 0, sizeof resmgr_attr);
  resmgr_attr.nparts_max = 1;
  resmgr_attr.msg_max_size = 2048;

  /* initialize functions for handling messages */
  iofunc_func_init (_RESMGR_CONNECT_NFUNCS, &connect_funcs,
		    _RESMGR_IO_NFUNCS, &io_funcs);

  io_funcs.devctl = io_devctl;	/* For handling _IO_DEVCTL, sent by devctl() */

  /* initialize attribute structure used by the device */
  iofunc_attr_init (&attr, S_IFNAM | 0666, 0, 0);

  /* attach our device name */
  id = resmgr_attach (dpp,	/* dispatch handle        */
		      &resmgr_attr,	/* resource manager attrs */
		      "/dev/hitec",	/* device name            */
		      _FTYPE_ANY,	/* open type              */
		      0,	/* flags                  */
		      &connect_funcs,	/* connect routines       */
		      &io_funcs,	/* I/O routines           */
		      &attr);	/* handle                 */
  if (id == -1)
    {
      fprintf (stderr, "%s: Unable to attach name.\n", argv[0]);
      return EXIT_FAILURE;
    }

  /* allocate a context structure */
  ctp = dispatch_context_alloc (dpp);

  /* gain access to IO priviledges and system clock */
  ThreadCtl (_NTO_TCTL_IO, 0);

  newcp.nsec = 1e9 / IRQ0_FREQ;
  newcp.fract = 0L;
  
  /* change system clock */
  if (ClockPeriod (CLOCK_REALTIME, &newcp, &oldcp, 0) == -1) {
      perror ("ClockPeriod()");
      return EXIT_FAILURE;
  } else {
#ifdef VARIANT_g
			printf("ClockPeriod: new = %ldns, old = %ldns\n",
							newcp.nsec, oldcp.nsec);
#endif
	}

  /* return to previous clock after Ctrl-C */
  if (signal(SIGINT, sigint) == SIG_ERR) {
	  perror("signal()");
  }
  
  /* gain access to LPT port */
  if (mmap_device_io(8, LPT) == MAP_DEVICE_FAILED) {
  	perror("mmap_device_io()");
  	return EXIT_FAILURE;
  }

  /* attach clock interrupt handler */
  irq_id = InterruptAttach(IRQ0, isr_handler, NULL, 0, _NTO_INTR_FLAGS_PROCESS);
  if (irq_id == -1) {
  	perror("InterruptAttach()");
  	return EXIT_FAILURE;
  }
  
  for (i = 0; i < 2; i++) {
  	printf("\a"); fflush(stdout);
  	delay(400);
  }

  pid = pthread_self();

  pthread_getschedparam(pid, &policy, &param);
#ifdef VARIANT_g  
  printf("pri=%d", param.sched_priority);
  switch(policy) {
		case SCHED_FIFO:     printf("f"); break;
		case SCHED_RR:       printf("r"); break;
		case SCHED_OTHER:    printf("o"); break;
		case SCHED_SPORADIC: printf("s"); break;
		default: printf("?");
  }
  printf("\n");
#endif /* VARIANT_g */

  policy = SCHED_FIFO; // SCHED_FIFO, SCHED_RR
  param.sched_priority = param.sched_priority+1;

#if 1
  if (pthread_setschedparam(pid, policy, &param) == -1) {
	  perror("sched_setscheduler()");
  }

#ifdef VARIANT_g
  pthread_getschedparam(pid, &policy, &param);
  printf("pri=%d", param.sched_priority);
  switch(policy) {
		case SCHED_FIFO:     printf("f"); break;
		case SCHED_RR:       printf("r"); break;
		case SCHED_OTHER:    printf("o"); break;
		case SCHED_SPORADIC: printf("s"); break;
		default: printf("?");
  }
  printf("\n");
#endif /* VARIANT_g */
#endif

  /* start the resource manager message loop */
  while (1)
    {
      if ((ctp = dispatch_block (ctp)) == NULL)
	{
	  if (errno != EINTR) {
	  	perror("dispatch_block()");
	  }
	  return EXIT_FAILURE;
	}
      dispatch_handler (ctp);
    }
}

int
pos2pwm(data_t *pos)
{
	int i, t;

#ifdef VARIANT_g	
	printf("val:");
	for (i = 0; i < MOTORS; i++) {
		printf(" %d", pos->tx_pos[i]);
	}
	printf("\n");
#endif /* VARIANT_g */
	
	for (t = 0; t < PWM_MOD; t++) {
		pwm_val[t] = 0x00;
		for (i = 0; i < MOTORS; i++) {
			pwm_val[t] |= (pos->tx_pos[i]*PWM_MOD/100 >= t) ? (1 << i) : 0x00;
		}
	}
	
	return t;
}					

int
io_devctl (resmgr_context_t * ctp, io_devctl_t * msg, RESMGR_OCB_T * ocb)
{
  int nbytes, status;
  union
  {
    data_t data;
    // ... other devctl types you can receive
  } *rx_data;
  data_t previous;

  /*
     Let common code handle DCMD_ALL_* cases.
     You can do this before or after you intercept devctl's depending
     on your intentions.  Here we aren't using any pre-defined values
     so let the system ones be handled first.
   */
  if ((status = iofunc_devctl_default (ctp, msg, ocb)) != _RESMGR_DEFAULT)
    {
      return (status);
    }
  status = nbytes = 0;
  
  /*
     Note this assumes that you can fit the entire data portion of
     the devctl into one message.  In reality you should probably
     perform a MsgReadv() once you know the type of message you
     have received to suck all of the data in rather than assuming
     it all fits in the message.  We have set in our main routine
     that we'll accept a total message size of up to 2k so we
     don't worry about it in this example where we deal with ints.
   */
  rx_data = _DEVCTL_DATA (msg->i);

  /*
     Three examples of devctl operations.
     SET: Setting a value (int) in the server
     GET: Getting a value (int) from the server
     SETGET: Setting a new value and returning with the previous value
   */
  switch (msg->i.dcmd)
    {
    case HITEC_DEVCTL_SETVAL:
      if (msg->i.nbytes != sizeof(data_t))
        return (EMSGSIZE);
      curr_pos = rx_data->data;
      pos2pwm(&curr_pos);
      nbytes = 0;
      break;

    case HITEC_DEVCTL_GETVAL:
      if (msg->i.nbytes != sizeof(data_t))
        return (EMSGSIZE);
      rx_data->data = curr_pos;
      nbytes = sizeof (data_t);
      break;

    case HITEC_DEVCTL_SETGET:
      if (msg->i.nbytes != sizeof(data_t))
        return (EMSGSIZE);
      previous = curr_pos;
      curr_pos = rx_data->data;
      pos2pwm(&curr_pos);
      rx_data->data = previous; //Overwrites tx data
      nbytes = sizeof(data_t);
      break;

    default:
      return (ENOSYS);
    }

  /* Clear the return message ... note we saved our data _after_ this */
  memset (&msg->o, 0, sizeof (msg->o));

  /*
     If you wanted to pass something different to the return
     field of the devctl() you could do it through this member.
   */
  msg->o.ret_val = status;

  /* Indicate the number of bytes and return the message */
  msg->o.nbytes = nbytes;
  return (_RESMGR_PTR (ctp, &msg->o, sizeof (msg->o) + nbytes));
}

const struct sigevent *
isr_handler (void *arg, int id)
{
	static unsigned int pulse = 0;
	
	if (pulse < PWM_MIN) {
		out8(LPT, 0xff);
	} else if (pulse < PWM_MAX) {
		out8(LPT, pwm_val[pulse - PWM_MIN]);
	} else {
		out8(LPT, 0x00);
	}	

	pulse++;
	pulse%=PWM_CYCLE;
    return (NULL);
}

